import inspect
import json
from functools import partial
from typing import Literal

from instructor import AsyncInstructor, Instructor
from loguru import logger
from pydantic import UUID4, BaseModel, create_model

from ayy import tools
from ayy.dialog import (
    Dialog,
    DialogToolSignature,
    MessagePurpose,
    ModelName,
    Task,
    Tool,
    add_dialog_message,
    add_task_message,
    assistant_message,
    create_creator,
    dialog_to_kwargs,
    get_last_message,
    user_message,
)
from ayy.dialogs import DIALOG_NAMER_DIALOG
from ayy.func_utils import function_to_type, get_function_info, get_functions_from_module
from ayy.torm import (
    add_task_tools,
    get_dialogs_with_signatures,
    get_next_task_tool,
    get_task_tools,
    load_dialog,
    load_task,
    save_dialog,
    save_task,
    toggle_task_tool_usage,
)

MODEL_NAME = ModelName.GEMINI_FLASH
PINNED_TOOLS = set(["ask_user", "call_ai"])
CONTINUE_DIALOG = True
DEFAULT_PROMPT = "Generate a response if you've been asked. Otherwise, ask the user how they are doing."
DEFAULT_TOOL = Tool(reasoning="", name="call_ai", prompt=DEFAULT_PROMPT)
TOOL_ATTEMPT_LIMIT = 3


class TaskQuery(BaseModel):
    task_query: str


async def handle_creator_error(task: Task, db_name: str, error: Exception, tool_name: str = ""):
    res = f"Whoops! Something went wrong. Here's the error:\n{error}"
    if tool_name:
        res = f"Whoops! Something went wrong in '{tool_name}'. Here's the error:\n{error}"
    await add_task_tools(
        task=task, tools=[Tool(reasoning=res, name="ask_user", prompt=res)], db_name=db_name, run_next=True
    )


def get_dialog_signature(dialog: Dialog) -> DialogToolSignature | None:
    namer = create_creator(model_name=DIALOG_NAMER_DIALOG.model_name)
    try:
        namer_res = namer.create(
            **dialog_to_kwargs(dialog=DIALOG_NAMER_DIALOG, messages=dialog.messages),
            response_model=DialogToolSignature,
        )
        logger.info(f"dialog signature: {namer_res}")
    except Exception:
        logger.exception("Error getting dialog signature")
        return None
    return None if namer_res.name == "" else namer_res  # type: ignore


async def run_ask_user(db_name: str, task: UUID4 | str | Task, dialog: UUID4 | str | Dialog, tool: Tool) -> Task:
    task = await load_task(task=task, db_name=db_name)
    dialog = await load_dialog(dialog=dialog, db_name=db_name)
    tool.prompt = tool.prompt or DEFAULT_PROMPT
    res = input(f"{tool.prompt}\n> ")
    task = add_task_message(task=task, message=assistant_message(content=tool.prompt))
    task = add_task_message(task=task, message=user_message(content=res))
    dialog = add_dialog_message(dialog=dialog, message=assistant_message(content=tool.prompt))
    dialog = add_dialog_message(dialog=dialog, message=user_message(content=res))
    await save_dialog(dialog=dialog, db_name=db_name)
    await save_task(task=task, db_name=db_name)
    return task


async def run_call_ai(
    db_name: str,
    task: UUID4 | str | Task,
    dialog: UUID4 | str | Dialog,
    tool: Tool,
    creator: Instructor | AsyncInstructor | None = None,
) -> Task:
    task = await load_task(task=task, db_name=db_name)
    dialog = await load_dialog(dialog=dialog, db_name=db_name)
    creator = creator or create_creator(model_name=dialog.model_name)
    assistant_message_purpose = (
        MessagePurpose.CONVO if tool.name in PINNED_TOOLS | {"get_selected_tools"} else MessagePurpose.TOOL
    )
    tool.prompt = tool.prompt or DEFAULT_PROMPT
    logger.info(f"adding user message: {tool.prompt}")
    tp_message = user_message(content=tool.prompt, purpose=MessagePurpose.TOOL)
    # task = add_task_message(task=task, message=tp_message)
    dialog = add_dialog_message(dialog=dialog, message=tp_message)
    logger.info(f"\n\nCalling AI with messages: {task.messages}\n\n")
    res = creator.create(**dialog_to_kwargs(dialog=dialog), response_model=str)
    logger.success(f"call_ai result: {res}")
    logger.info(f"adding assistant message: {res}")
    res_message = assistant_message(content=res, purpose=assistant_message_purpose)
    if assistant_message_purpose != MessagePurpose.TOOL:
        task = add_task_message(task=task, message=res_message)
    dialog = add_dialog_message(dialog=dialog, message=res_message)
    await save_dialog(dialog=dialog, db_name=db_name)
    await save_task(task=task, db_name=db_name)
    return task


async def get_selected_tools(
    db_name: str, task: UUID4 | str | Task, dialog: UUID4 | str | Dialog, selected_tools: list[Tool]
) -> Task:
    "Get and push a list of selected tools for the task"
    task = await load_task(task=task, db_name=db_name)
    dialog = await load_dialog(dialog=dialog, db_name=db_name)
    tools_str = "\n".join([f"Tool {i}:\n{tool}" for i, tool in enumerate(selected_tools, start=1)])
    st_message = assistant_message(
        f"<selected_tools>\n{tools_str}\n</selected_tools>", purpose=MessagePurpose.TOOL
    )
    # task = add_task_message(task=task, message=st_message)
    dialog = add_dialog_message(dialog=dialog, message=st_message)
    await save_dialog(dialog=dialog, db_name=db_name)
    await save_task(task=task, db_name=db_name)
    await add_task_tools(task=task, tools=selected_tools, db_name=db_name)
    return task


async def run_dialog_as_tool(db_name: str, dialog: UUID4 | str | Dialog, task_query: str) -> Task:
    task = await new_task(db_name=db_name, dialog=dialog, task_query=task_query, continue_dialog=False)
    return task


async def run_tool(
    db_name: str,
    task: UUID4 | str | Task,
    dialog: UUID4 | str | Dialog,
    tool: Tool,
    tool_is_dialog: bool = False,
    available_tools: list[str] | set[str] | None = None,
    creator: Instructor | AsyncInstructor | None = None,
    ignore_default_values: bool = False,
    skip_default_params: bool = False,
) -> Task:
    task = await load_task(task=task, db_name=db_name)
    dialog = await load_dialog(dialog=dialog, db_name=db_name)
    is_async = False
    if tool_is_dialog:
        selected_tool = partial(run_dialog_as_tool, db_name, await load_dialog(dialog=tool.name, db_name=db_name))
        is_async = True
    else:
        try:
            selected_tool = getattr(tools, tool.name, globals().get(tool.name, None))
            if selected_tool is None:
                raise ValueError(f"Tool '{tool.name}' not found in tools module or current module")
            if not inspect.isfunction(selected_tool):
                raise ValueError(
                    f"Tool '{tool.name}' is not a function.\nGot {type(selected_tool).__name__} instead"
                )
            is_async = inspect.iscoroutinefunction(selected_tool)
        except AttributeError:
            raise ValueError(f"Tool '{tool.name}' not found in tools module")
    if tool.prompt:
        logger.info(f"adding user message: {tool.prompt}")
        message_purpose = MessagePurpose.CONVO if tool.name == "get_selected_tools" else MessagePurpose.TOOL
        tp_message = user_message(content=tool.prompt, purpose=message_purpose)
        if message_purpose != MessagePurpose.TOOL:
            task = add_task_message(task=task, message=tp_message)
        dialog = add_dialog_message(dialog=dialog, message=tp_message)
        await save_dialog(dialog=dialog, db_name=db_name)
        await save_task(task=task, db_name=db_name)
    tool_type = (
        getattr(tools, "tool_types", globals().get("tool_types", {})).get(tool.name, None)
        if not tool_is_dialog
        else TaskQuery
    )
    all_tools = available_tools or [tool_member[0] for tool_member in get_functions_from_module(module=tools)]
    if tool.name == "get_selected_tools" and len(all_tools) > 0:
        selected_tool = partial(get_selected_tools, db_name, task, dialog)
        tool_type = list[
            create_model(
                "SelectedTool",
                name=(Literal[*all_tools], ...),  # type: ignore
                __base__=Tool,
            )
        ]
        is_async = True
    if tool_type is None:
        tool_type = function_to_type(
            func=selected_tool,
            ignore_default_values=ignore_default_values,
            skip_default_params=skip_default_params,
        )
    # logger.info(f"\n\nCalling {tool.name} with tool_type: {tool_type}\n\n")
    creator = creator or create_creator(model_name=dialog.model_name)
    creator_res = creator.create(
        **dialog_to_kwargs(dialog=dialog),
        response_model=tool_type,  # type: ignore
    )
    logger.info(f"{tool.name} creator_res: {creator_res}")
    if isinstance(creator_res, BaseModel):
        if is_async:
            res = await selected_tool(**creator_res.model_dump())
        else:
            res = selected_tool(**creator_res.model_dump())
    elif is_async:
        res = await selected_tool(creator_res)  # type: ignore
    else:
        res = selected_tool(creator_res)  # type: ignore
    logger.success(f"{tool.name} result: {res}")
    if isinstance(res, Task):
        if res.name in [dialog.name, task.name]:
            return res
        res = get_last_message(messages=res.messages, role="assistant")
        if res is not None:
            res = res["content"]
    if res is not None:
        logger.info(f"adding assistant message: {res}")
        res_message = assistant_message(content=str(res), purpose=MessagePurpose.TOOL)
        # task = add_task_message(task=task, message=res_message)
        dialog = add_dialog_message(dialog=dialog, message=res_message)
    await save_dialog(dialog=dialog, db_name=db_name)
    await save_task(task=task, db_name=db_name)
    return task


async def run_selected_tool(
    db_name: str,
    task: UUID4 | str | Task,
    dialog: UUID4 | str | Dialog,
    tool: Tool,
    tool_is_dialog: bool = False,
    available_tools: list[str] | set[str] | None = None,
    creator: Instructor | AsyncInstructor | None = None,
) -> Task:
    task = await load_task(task=task, db_name=db_name)
    dialog = await load_dialog(dialog=dialog, db_name=db_name)
    creator = creator or create_creator(model_name=dialog.model_name)
    if tool.name.lower() == "ask_user":
        task = await run_ask_user(db_name=db_name, task=task, dialog=dialog, tool=tool)
    elif tool.name.lower() == "call_ai":
        task = await run_call_ai(db_name=db_name, task=task, dialog=dialog, tool=tool, creator=creator)
    else:
        task = await run_tool(
            db_name=db_name,
            task=task,
            dialog=dialog,
            tool=tool,
            tool_is_dialog=tool_is_dialog,
            available_tools=available_tools,
            creator=creator,
        )
    return task


async def run_tools(
    db_name: str,
    task: UUID4 | str | Task,
    dialog: UUID4 | str | Dialog,
    creator: Instructor | AsyncInstructor | None = None,
    available_tools: list[str] | set[str] | None = None,
    dialogs: list[str] | set[str] | None = None,
    continue_dialog: bool = CONTINUE_DIALOG,
) -> Task:
    task = await load_task(task=task, db_name=db_name)
    dialog = await load_dialog(dialog=dialog, db_name=db_name)
    creator = creator or create_creator(model_name=dialog.model_name)
    while True:
        next_tool = await get_next_task_tool(task=task, db_name=db_name)
        logger.info(f"next_tool: {next_tool}")
        if next_tool is None:
            break
        attempts = 0
        while attempts < TOOL_ATTEMPT_LIMIT:
            try:
                task = await run_selected_tool(
                    db_name=db_name,
                    task=task,
                    creator=creator,
                    dialog=dialog,
                    tool=next_tool.tool,
                    tool_is_dialog=next_tool.tool.name in dialogs if dialogs else False,
                    available_tools=available_tools,
                )
                await toggle_task_tool_usage(task_tool_id=next_tool.id, db_name=db_name)
                break
            except Exception as e:
                logger.exception(f"Error running tool '{next_tool.tool.name}'")
                await handle_creator_error(task=task, db_name=db_name, error=e, tool_name=next_tool.tool.name)
                attempts += 1
    used_tools = await get_task_tools(task=task, db_name=db_name, used=True)
    last_used_tool = used_tools[-1].tool if used_tools else DEFAULT_TOOL
    if continue_dialog:
        seq = int(last_used_tool.name == "ask_user")
        while True:
            if seq % 2 == 0 or last_used_tool.name == "call_ai":
                user_input = input("('q' or 'exit' or 'quit' to quit) > ")
                if user_input.lower().strip() in ["q", "exit", "quit"]:
                    return task
                task = await new_task(
                    db_name=db_name,
                    dialog=dialog,
                    task_query=user_input,
                    available_tools=available_tools,
                    continue_dialog=False,
                )
            else:
                await add_task_tools(task=task, tools=DEFAULT_TOOL, db_name=db_name, used=True)
                try:
                    res = creator.create(**dialog_to_kwargs(dialog=dialog), response_model=str)
                    logger.success(f"ai response: {res}")
                    res_message = assistant_message(content=res)
                except Exception as e:
                    logger.exception("Error calling AI after continuing dialog")
                    res_message = assistant_message(
                        content=f"Whoops! Something went wrong. Here's the error:\n{e}",
                        purpose=MessagePurpose.ERROR,
                    )
                logger.info(f"adding message: {res_message['content']}")
                task = add_task_message(task=task, message=res_message)
                dialog = add_dialog_message(dialog=dialog, message=res_message)
                await save_dialog(dialog=dialog, db_name=db_name)
                await save_task(task=task, db_name=db_name)
            seq += 1

    logger.success(f"Messages: {task.messages[-2:]}")
    dialog_signature = get_dialog_signature(dialog=dialog)
    await save_dialog(dialog=dialog, db_name=db_name, dialog_tool_signature=dialog_signature)
    await save_task(task=task, db_name=db_name)
    return task


async def new_task(
    db_name: str,
    dialog: UUID4 | str | Dialog,
    task_query: str,
    task_name: str = "",
    creator: Instructor | AsyncInstructor | None = None,
    available_tools: list[str] | set[str] | None = None,
    recommended_tools: dict[int, str] | None = None,
    continue_dialog: bool = CONTINUE_DIALOG,
) -> Task:
    dialog = await load_dialog(dialog=dialog, db_name=db_name)
    task = Task(name=task_name, dialog_id=dialog.id)
    available_tools = (set(available_tools or []) | set(dialog.available_tools or [])) or []
    dialogs = await get_dialogs_with_signatures(db_name=db_name)
    dialog_names = []
    dialogs_as_tools = []
    for d in dialogs:
        if d.name != dialog.name:
            dialog_names.append(d.name)
            dialogs_as_tools.append(f"Tool:\n{d.dialog_tool_signature}")
    tool_names = []
    tools_list = []
    for _, func in get_functions_from_module(module=tools):
        if not available_tools or func.__name__ in set(available_tools) | PINNED_TOOLS:
            tool_names.append(func.__name__)
            tools_list.append(f"Tool:\n{get_function_info(func)}")
    tools_info = "\n\n".join(tools_list + dialogs_as_tools)
    av_message = user_message(content=f"Available tools for this task:\n{tools_info}", purpose=MessagePurpose.TOOL)
    # task = add_task_message(task=task, message=av_message)
    dialog = add_dialog_message(dialog=dialog, message=av_message)
    if recommended_tools is None:
        used_tools = await get_task_tools(task=task, db_name=db_name, used=True)
        recommended_tools = {tool.position: tool.tool.name for tool in used_tools}
    if recommended_tools:
        rec_message = user_message(
            content=f"Recommended tools for this task (in order):\n{json.dumps(recommended_tools, indent=2)}",
            purpose=MessagePurpose.TOOL,
        )
        # task = add_task_message(task=task, message=rec_message)
        dialog = add_dialog_message(dialog=dialog, message=rec_message)
    await save_dialog(dialog=dialog, db_name=db_name)
    await save_task(task=task, db_name=db_name)
    await add_task_tools(
        task=task, tools=[Tool(reasoning="", name="get_selected_tools", prompt=task_query)], db_name=db_name
    )
    return await run_tools(
        db_name=db_name,
        task=task,
        dialog=dialog,
        creator=creator,
        available_tools=set(tool_names) | set(dialog_names) | PINNED_TOOLS,
        dialogs=set(dialog_names),
        continue_dialog=continue_dialog,
    )



from tortoise import fields, models

from ayy.dialog import MAX_MESSAGE_TOKENS, MAX_TOKENS, MODEL_NAME, TEMPERATURE

DEFAULT_APP_NAME = "tasks"


class Dialog(models.Model):
    id = fields.UUIDField(pk=True)
    name = fields.CharField(max_length=255, default="")
    system = fields.TextField(default="")
    messages = fields.JSONField(default=list)
    model_name = fields.CharField(max_length=255, default=MODEL_NAME.value)
    max_message_tokens = fields.IntField(default=MAX_MESSAGE_TOKENS)
    creation_config = fields.JSONField(default=dict(temperature=TEMPERATURE, max_tokens=MAX_TOKENS))
    dialog_tool_signature = fields.JSONField(default=dict)
    available_tools = fields.JSONField(default=list)
    include_tool_guidelines = fields.BooleanField(default=True)

    class Meta:  # type: ignore
        app = DEFAULT_APP_NAME
        table = "dialog"


# class DialogTool(models.Model):
#     id = fields.IntField(pk=True)
#     dialog = fields.ForeignKeyField(f"{DEFAULT_APP_NAME}.Dialog", related_name="dialog_tool")
#     position = fields.IntField()
#     reasoning = fields.TextField()
#     name = fields.CharField(max_length=255)
#     prompt = fields.TextField()
#     used = fields.BooleanField(default=False)
#     timestamp = fields.DatetimeField(auto_now_add=True)

#     class Meta:  # type: ignore
#         app = DEFAULT_APP_NAME
#         table = "dialog_tool"
#         ordering = ["dialog_id", "position"]


class Task(models.Model):
    id = fields.UUIDField(pk=True)
    name = fields.CharField(max_length=255, default="")
    dialog = fields.ForeignKeyField(f"{DEFAULT_APP_NAME}.Dialog", related_name="task")
    messages = fields.JSONField(default=list)

    class Meta:  # type: ignore
        app = DEFAULT_APP_NAME
        table = "task"


class TaskTool(models.Model):
    id = fields.IntField(pk=True)
    task = fields.ForeignKeyField(f"{DEFAULT_APP_NAME}.Task", related_name="task_tool")
    position = fields.IntField()
    reasoning = fields.TextField()
    name = fields.CharField(max_length=255)
    prompt = fields.TextField()
    used = fields.BooleanField(default=False)
    timestamp = fields.DatetimeField(auto_now_add=True)

    class Meta:  # type: ignore
        app = DEFAULT_APP_NAME
        table = "task_tool"
        ordering = ["task_id", "position"]



from ayy.dialog import Dialog, ModelName, exchange
from ayy.memory import SemanticCategory
from ayy.prompts import NAME_DIALOG, SUMMARIZE_MESSAGES

DIALOG_NAMER_DIALOG = Dialog(model_name=ModelName.GEMINI_FLASH, system=NAME_DIALOG, name="dialog_namer_dialog")


SUMMARIZER_DIALOG = Dialog(
    model_name=ModelName.GEMINI_FLASH,
    system=SUMMARIZE_MESSAGES
    + f"\nAvailable categories for semantic memories: {SemanticCategory._member_names_}",
    messages=[
        *exchange(user="My name is Hamza", assistant="Hey Hamza, how can I help you today?"),
        *exchange(
            user="I'm planning a trip to visit my family in New York. How's the weather there?",
            assistant="Let me check the weather for you. It looks like it's currently sunny in New York with temperatures around 75°F (24°C). The forecast shows similar weather for the next few days.",
        ),
        *exchange(
            user="I love sunny days! That's perfect for my trip.",
            assistant="That's great that the weather aligns with your preferences! Would you like me to help you book a flight to New York?",
        ),
        *exchange(
            user="Yes please, I'd prefer a morning flight if possible",
            assistant="I'll look for morning flights. Do you have any specific dates in mind for your travel?",
        ),
        *exchange(
            user="Next weekend would be ideal",
            assistant="I found several morning flights for next weekend. There's a direct flight on Saturday at 8:30 AM, arriving at 11:45 AM. Would that work for you?",
        ),
        *exchange(
            user="That sounds perfect",
            assistant="Great! I've booked the flight for you. Here are the details:\n- Flight AA1234\n- Saturday, 8:30 AM departure\n- Direct flight to JFK\n- Confirmation code: XYZ789",
        ),
        *exchange(
            user="Thanks! Could you remind me what time I arrive?",
            assistant="Your flight arrives at JFK at 11:45 AM. Would you like me to help arrange transportation from the airport?",
        ),
        *exchange(
            user="No thanks, my family will pick me up",
            assistant="Perfect! Let me know if you need anything else regarding your trip.",
        ),
        *exchange(
            user="Summarize the conversation so far.",
            assistant="""Summary: {
                "bullet_points": [
                    "Trip planning initiated to visit family in New York",
                    "Flight booked: Saturday morning direct flight to JFK",
                    "Airport pickup arranged with family"
                ],
                "semantic_memories": [
                    {
                        "name": "personal_info",
                        "content": "User's name is Hamza",
                        "category": "identity",
                        "confidence": 1.0,
                    },
                    {
                        "name": "family_location",
                        "content": "Has family in New York",
                        "category": "relationships",
                        "confidence": 1.0,
                    },
                    {
                        "name": "travel_preferences",
                        "content": "Prefers sunny weather and morning flights",
                        "category": "preferences",
                        "confidence": 0.8,
                    }
                ],
                "episodic_memories": [
                    {
                        "name": "flight_booking",
                        "content": {
                            "flight_number": "AA1234",
                            "departure": "8:30 AM Saturday",
                            "arrival": "11:45 AM",
                            "confirmation": "XYZ789"
                        },
                        "context": "current_trip",
                        "confidence": 1.0,
                    },
                    {
                        "name": "airport_pickup",
                        "content": "Family will provide airport pickup",
                        "context": "current_trip",
                        "confidence": 1.0,
                    }
                ]
            }""",
        ),
    ],
    name="summarizer_dialog",
)



from datetime import datetime
from functools import partial

import instructor
from google.generativeai import GenerativeModel
from loguru import logger
from sqlmodel import Field, Relationship, Session, SQLModel, create_engine

from ayy.dialog import Content, ModelName, assistant_message, chat_message, messages_to_kwargs, user_message

MODEL = ModelName.GEMINI_FLASH


class Dialog(SQLModel, table=True):
    id: int | None = Field(default=None, primary_key=True)
    system: str = ""
    model_name: ModelName = MODEL
    messages: list["Message"] = Relationship(back_populates="dialog")


class Message(SQLModel, table=True):
    id: int | None = Field(default=None, primary_key=True)
    dialog_id: int = Field(foreign_key="dialog.id")
    role: str
    content: str
    created_at: datetime = Field(default_factory=datetime.now)
    dialog: Dialog = Relationship(back_populates="messages")


def add_assistant_message(dialog: Dialog, creator: Content) -> Dialog:
    try:
        res = (
            creator(
                **messages_to_kwargs(
                    messages=[chat_message(role=msg.role, content=msg.content) for msg in dialog.messages],
                    system=dialog.system,
                    model_name=dialog.model_name,
                )
            )
            if callable(creator)
            else creator
        )
    except Exception as e:
        logger.exception(f"Error in respond. Last message: {dialog.messages[-1]}")
        res = f"Error: {e}"
    dialog.messages.append(Message(**assistant_message(content=res)))
    dialog.updated_at = datetime.now()
    return dialog


def add_user_message(dialog: Dialog, content: Content, template: Content = "") -> Dialog:
    dialog.messages.append(Message(**user_message(content=content, template=template)))
    dialog.updated_at = datetime.now()
    return dialog


creator = partial(
    instructor.from_gemini(client=GenerativeModel(model_name=MODEL), mode=instructor.Mode.GEMINI_JSON).create,
    response_model=str,
)

sqlite_file_name = "database.db"
sqlite_url = f"sqlite:///{sqlite_file_name}"

engine = create_engine(sqlite_url, echo=True)


SQLModel.metadata.create_all(engine)

dialog = Dialog(
    system="Talk like a pirate.",
    messages=[
        Message(**user_message("thoughts on messi?")),
        Message(**assistant_message("arr! he be amazing cap'n")),
    ],
)

with Session(engine) as session:
    seq = 0
    while True:
        session.add(dialog)
        session.commit()
        session.refresh(dialog)
        if seq % 2 == 0:
            user_input = input("> ")
            if user_input.lower() == "q":
                break
            dialog = add_user_message(dialog=dialog, content=user_input)
        else:
            dialog = add_assistant_message(dialog=dialog, creator=creator)
        seq += 1



import asyncio
from typing import Any, Literal, Set
from urllib.parse import urljoin, urlparse

from crawl4ai import AsyncWebCrawler
from loguru import logger

MAX_DEPTH = 2
MAX_LINKS = 5


def call_ai(inputs: Any) -> Any:
    "Not a pre-defined tool."
    return inputs


def ask_user(prompt: str) -> str:
    "Prompt the user for a response"
    return prompt


def get_weather(
    day: Literal["Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday"], location: str
) -> str:
    "get the weather at a day in a location"
    if day == "Monday" and location.lower() == "blackpool":
        return "It's raining"
    elif day == "Tuesday" and location.lower() == "london":
        return "It's sunny"
    elif day == "Wednesday" and location.lower() == "manchester":
        return "It's cloudy"
    else:
        return "It's overcast"


def list_available_grounds(location: str) -> list[str]:
    "list all available grounds in a location"
    if location.lower() == "blackpool":
        return ["The Hawthorns", "The Den", "The New Den"]
    elif location.lower() == "london":
        return ["Wembley Stadium", "Emirates Stadium", "Tottenham Hotspur Stadium"]
    elif location.lower() == "manchester":
        return ["The Old Trafford", "The Etihad Stadium", "The Maine Road"]
    else:
        return ["Palm Football"]


def download_video(url: str) -> str:
    "download video from url and return the local path"
    return f"videos/{url.split('/')[-1]}"


async def _recursive_crawl(
    url: str,
    max_depth: int = MAX_DEPTH,
    max_links: int = MAX_LINKS,
    same_domain_only: bool = True,
    prefixes: list[str] | None = None,
) -> dict:
    """
    Recursively crawl starting from a URL up to a specified depth.

    Args:
        url: The URL to start crawling from
        max_depth: Maximum depth of recursion (default: 2)
        max_links: Maximum number of links to follow (default: 100)
        same_domain_only: Only follow links within the same domain (default: True)
        prefixes: List of prefixes to follow (default: None). So only links starting with these prefixes will be followed. If same_domain_only is True, it will be automatically added.
    """
    visited: Set[str] = set()
    results = {}
    start_domain = urlparse(url).netloc
    prefixes = prefixes or []
    if same_domain_only:
        prefixes = [p for p in prefixes + [start_domain] if urlparse(p).netloc == start_domain]
    logger.info(f"Crawling {url} with prefixes {prefixes}")

    async def crawl_url(url: str, depth: int):
        logger.info(f"Crawling {url} at depth {depth} with netloc {urlparse(url).netloc}")
        if depth > max_depth or url in visited or len(results) > max_links:
            return

        if prefixes and not any(url.startswith(p) for p in prefixes):
            logger.warning(f"Skipping {url} because it is not in prefixes")
            return

        visited.add(url)

        async with AsyncWebCrawler(verbose=True) as crawler:
            try:
                result = await crawler.arun(url=url)
                results[url] = result.markdown

                # Extract links from the page
                if result.links and depth < max_depth:
                    for _, links in result.links.items():
                        for link in links:
                            if "href" in link:
                                next_url = urljoin(url, link["href"])
                                await crawl_url(url=next_url, depth=depth + 1)
            except Exception:
                logger.exception(f"Error crawling {url}")

    await crawl_url(url, 1)
    return results


def crawl_url(
    url: str,
    max_depth: int = MAX_DEPTH,
    max_links: int = MAX_LINKS,
    same_domain_only: bool = True,
    prefixes: list[str] | None = None,
) -> dict:
    """
    Recursively crawl starting from a URL up to a specified depth.

    Args:
        url: The URL to start crawling from
        max_depth: Maximum depth of recursion (default: 2)
        max_links: Maximum number of links to follow (default: 100)
        same_domain_only: Only follow links within the same domain (default: True)
        prefixes: List of prefixes to follow (default: None). So only links starting with these prefixes will be followed. If same_domain_only is True, it will be automatically added.
    """
    return asyncio.run(
        _recursive_crawl(
            url=url, max_depth=max_depth, max_links=max_links, same_domain_only=same_domain_only, prefixes=prefixes
        )
    )



SELECT_TOOLS = """
You have a list tools at your disposal. Each tool is a function with a signature and optional docstring.
Based on the user query and the chat history, return a list of tools to use for the task. The tools will be used in that sequence.
You can assume that a tool would have access to the result of a previous tool call.
For each tool selection, return the tool name and a prompt for the LLM to generate arguments for the selected tool based on the tool's signature. Make sure to not forget any parameters. Don't mention other tools in the prompt. The LLM will receive the messages so far and the tools calls and results up until that point. If the tool doesn't have any parameters, then it doesn't need a prompt.
Remember the actual user query/task throughout your tool selection process. Especially when creating the prompt for the LLM.
More often than not, the last tool would be 'call_ai' to generate the final response.
Pay close attention to the information you do have and the information you do not have. Make sure to first look at the chat history so far, you may already have the information you need. If you don't have the information, ask the user for it. Don't make assumptions. Even if you think the user does not have it, just talk it out with the user. DO NOT MAKE STUFF UP.
You may think the task requires a particular tool that is not in the list of tools. If so, clearly let the user know using 'call_ai' or 'ask_user'. You could even suggest a tool that might be useful.

When to use 'ask_user':
    - To ask the user something.
    - When the right tool is obvious but you need some extra data based on the function signature, then select the ask_user tool before selecting the actual tool and in your prompt explicitly state the extra information you need. So the tool should still be selected, you're just pairing it with a prior ask_user call.
    - A tool may have some default values. But if you think they should be provided by the user for the current task, ask for them.

Don't use the 'ask_user' tool to ask the user to fix obvious typos, do that yourself. That's a safe assumption. People make typos all the time.

When to use 'call_ai':
    - Whenever you want. In between or at the start/end.
    - To get the AI to generate something. This could be the final response or something in between tool calls.
    - To extract information before the next tool call.
"""


NAME_DIALOG = """
The tool names are descriptive. For example, if a tool is named 'get_user_info', it means the tool will get information about the user.
Throughout the conversation, you would have selected and used some tools to perform the task. If you had to encapsulate this whole flow in a single tool call, what would the function signature look like? With a verbose name and detailed docstring? In a JSON format.
TAKE THE WHOLE CONVERSATION AND WORKFLOW INTO ACCOUNT.

The function should always have just one parameter: `task_query: str`

In the docstring, list each tool that will be used in the flow, including:
- The tool's full signature
- A brief explanation of why/when this tool is needed
- Any important considerations about its usage

The name and docstring should still be generic. For example, if the user had a task and then they defined a specific day or location, don't include that specific day or location in the name/docstring.
Also create a system prompt based on the tools used + workflow + conversation history. It would be similar to the docstring but with more details and guidelines and framed like instructions for an LLM on how to solve this or similar tasks. Address the LLM as 'you' in the system prompt.
If multiple tools are used, incorporate them intuitively into the name using _and_ or _or_ etc.
If you just used 1-2 tools (excluding 'ask_user' and 'call_ai'), then there is no need to encapsulate the whole flow in a single tool call. Because that would clash with the actual tools and cause confusion.
For example, if you had just used the "set_reminder" tool in the workflow, and you decide to name this workflow "set_reminder_for_user", then which one would be selected the next time? There's no point becase a single tool call isn't a workflow. It's better to just use the tool. In that case, just return empty strings like this: {"name": "", "signature": "", "docstring": "", "system": ""}

For example:

Task from the user: "I want to buy a new phone"

Tools selected and workflow:
    - get_user_info() -> dict
    - get_user_location(user_id: str) -> str
    - get_user_budget(user_id: str) -> float
    - get_user_preferences(user_id: str, category: str = "phones") -> list[str]
    - check_phone_availability(location: str, preferences: list[str], budget: float) -> list[dict]
    - buy_phone(user_id: str, phone_id: str, payment_method: str) -> dict
    - call_ai

Signature (generated by you):
    {
        "name": "get_user_preferences_and_check_phone_availability_and_buy_phone",
        "signature": "get_user_preferences_and_check_phone_availability_and_buy_phone(task_query: str)",
        "docstring": '''Handles the complete phone purchase workflow based on user preferences and constraints.

        Tools used:
        1. get_user_info() -> dict
           - Retrieves basic user information needed for subsequent calls
           - Must be called first to get user_id

        2. get_user_location(user_id: str) -> str
           - Gets user's location for local availability checking
           - Required to ensure phones are available in user's area

        3. get_user_budget(user_id: str) -> float
           - Determines user's budget constraints
           - Used to filter available phone options

        4. get_user_preferences(user_id: str, category: str = 'phones') -> list[str]
           - Collects specific phone preferences (brand, features, etc.)
           - Helps narrow down phone selection

        5. check_phone_availability(location: str, preferences: list[str], budget: float) -> list[dict]
           - Searches for phones matching all criteria
           - Returns only phones that are actually available

        6. buy_phone(user_id: str, phone_id: str, payment_method: str) -> dict
           - Executes the final purchase
           - Called only after user confirms selection''',
        "system": '''You are a phone purchase assistant that helps users find and buy phones that match their preferences and constraints.

        Follow these steps:
        1. First get basic user information to establish context
        2. Determine user's location to check local availability
        3. Get budget constraints to filter options
        4. Collect detailed preferences about desired phone features
        5. Search for available phones matching all criteria
        6. Present options to user and help with final purchase

        Important guidelines:
        - Always verify budget before showing options
        - Check local availability before suggesting phones
        - Get explicit confirmation before purchase
        - Consider user's previous phone experiences
        - Explain key features of recommended phones
        - Highlight pros/cons of each option'''
    }

---

### Additional Examples:

**Example 1:**

**Task from the user:** "I need to organize my weekly meals"

**Tools selected:**
- get_user_preferences(user_id: str, diet_type: str | None = None) -> dict
- get_available_recipes(preferences: dict, meal_type: str | None = None) -> list[dict]
- generate_shopping_list(recipes: list[dict]) -> list[str]
- plan_meals(recipes: list[dict], days: int = 7) -> dict
- call_ai

**Signature (generated by you):**
{
    "name": "get_available_recipes_and_generate_shopping_list_and_plan_meals",
    "signature": "get_available_recipes_and_generate_shopping_list_and_plan_meals(task_query: str)",
    "docstring": '''Organizes weekly meals based on user preferences and dietary requirements.

    Tools used:
    1. get_user_preferences(user_id: str, diet_type: str | None = None) -> dict
       - Retrieves dietary restrictions and preferences
       - Essential for personalized meal planning

    2. get_available_recipes(preferences: dict, meal_type: str | None = None) -> list[dict]
       - Finds recipes matching user's dietary needs
       - Ensures variety in meal options

    3. generate_shopping_list(recipes: list[dict]) -> list[str]
       - Creates consolidated shopping list
       - Optimizes grocery shopping process

    4. plan_meals(recipes: list[dict], days: int = 7) -> dict
       - Arranges selected recipes into a weekly schedule
       - Considers balanced nutrition throughout the week''',
    "system": '''You are a meal planning assistant that helps users organize their weekly meals efficiently.

    Follow these steps:
    1. Gather dietary preferences and restrictions
    2. Find suitable recipes that match preferences
    3. Create an optimized shopping list
    4. Develop a balanced weekly meal plan

    Important guidelines:
    - Consider nutritional balance across the week
    - Account for dietary restrictions strictly
    - Suggest variety in meals to avoid repetition
    - Optimize shopping list to minimize waste
    - Group similar ingredients for shopping efficiency
    - Consider meal prep opportunities
    - Plan for leftovers when appropriate'''
}

---

**Example 2:**

**Task from the user:** "I want to track my daily expenses"

**Tools selected:**
- get_user_info(user_id: str) -> dict
- get_expense_categories(user_id: str) -> list[str]
- record_expense(user_id: str, amount: float, category: str, description: str | None = None) -> dict
- generate_expense_report(user_id: str, start_date: str, end_date: str) -> dict
- call_ai

**Signature (generated by you):**
{
    "name": "get_expense_categories_and_record_expense_and_generate_expense_report",
    "signature": "get_expense_categories_and_record_expense_and_generate_expense_report(task_query: str)",
    "docstring": '''Manages expense tracking and reporting workflow.

    Tools used:
    1. get_user_info(user_id: str) -> dict
       - Retrieves user's financial profile and preferences
       - Required for personalized expense tracking

    2. get_expense_categories(user_id: str) -> list[str]
       - Fetches available expense categories
       - Ensures consistent categorization of expenses

    3. record_expense(user_id: str, amount: float, category: str, description: str | None = None) -> dict
       - Logs individual expense entries
       - Core functionality for expense tracking

    4. generate_expense_report(user_id: str, start_date: str, end_date: str) -> dict
       - Creates summary reports of expenses
       - Provides insights into spending patterns''',
    "system": '''You are an expense tracking assistant that helps users monitor and analyze their spending.

    Follow these steps:
    1. Get user's financial profile
    2. Ensure proper expense categorization
    3. Record expenses accurately
    4. Generate insightful reports

    Important guidelines:
    - Maintain consistent categorization
    - Prompt for detailed descriptions when needed
    - Flag unusual spending patterns
    - Suggest budget optimizations
    - Provide periodic spending summaries
    - Compare expenses across time periods
    - Identify potential savings opportunities'''
}

---

**Example 3:**

**Task from the user:** "Help me prepare for a job interview"

**Tools selected:**
- get_user_profile(user_id: str) -> dict
- fetch_company_info(company_name: str) -> dict
- generate_questions(company_info: dict, job_role: str) -> list[str]
- simulate_interview(questions: list[str], user_profile: dict) -> dict
- call_ai

**Signature (generated by you):**
{
    "name": "get_company_info_and_generate_questions_and_simulate_interview",
    "signature": "get_company_info_and_generate_questions_and_simulate_interview(task_query: str)",
    "docstring": '''Provides comprehensive job interview preparation assistance.

    Tools used:
    1. get_user_profile(user_id: str) -> dict
       - Retrieves user's professional background
       - Essential for personalizing interview preparation

    2. fetch_company_info(company_name: str) -> dict
       - Gathers detailed information about the target company
       - Helps understand company culture and expectations

    3. generate_questions(company_info: dict, job_role: str) -> list[str]
       - Creates relevant interview questions
       - Based on company information and role requirements

    4. simulate_interview(questions: list[str], user_profile: dict) -> dict
       - Conducts mock interview sessions
       - Provides practice opportunities and feedback''',
    "system": '''You are an interview preparation assistant that helps candidates prepare for job interviews.

    Follow these steps:
    1. Review candidate's professional background
    2. Research target company thoroughly
    3. Generate relevant interview questions
    4. Conduct realistic mock interviews

    Important guidelines:
    - Tailor questions to job role and level
    - Focus on company-specific preparation
    - Provide constructive feedback
    - Practice both technical and behavioral questions
    - Suggest improvement areas
    - Help develop STAR method responses
    - Cover common and role-specific scenarios'''
}
"""
SUMMARIZE_MESSAGES = """
You are a skilled conversation summarizer. Your task is to analyze a conversation and create a concise, high-level summary that significantly reduces the message volume while preserving essential information.

The main purpose of summarization is to condense lengthy conversations into a brief, digestible format. Your summary should be much shorter than the original conversation - aim to capture key points and outcomes rather than listing every exchange. The conversation might already have a summary message at the start with a <summary_of_our_previous_conversation(s)> tag, that means that it was summarized before and has once again gotten too long. Make sure to incorporate that existing summary message into your summary.

Create a structured summary that includes:

1. Key bullet points that capture the essential flow and important moments of the conversation:
   - Focus on outcomes and decisions rather than back-and-forth exchanges
   - Combine related exchanges into single points
   - Skip pleasantries and redundant information
   - Highlight major state changes or decisions

2. Semantic memories that should be remembered long-term, such as:
   - Permanent facts about participants
   - Personal information
   - Long-term preferences and traits
   - Important rules or instructions
   - User feedback for improvement
   - Any other unchanging, crucial information

3. Episodic memories relevant to ongoing tasks:
   - Current context and parameters
   - Intermediate steps or results
   - Short-term preferences or needs
   - Time-sensitive details
   - Recent interactions or decisions

Focus on extracting meaningful insights while distinguishing between permanent and temporary information. Be concise but comprehensive.

Format your response as a structured Summary object with bullet_points, semantic_memories, and episodic_memories fields.
"""



from loguru import logger
from pydantic import UUID4, BaseModel
from tortoise import Tortoise, connections
from tortoise.contrib.pydantic import pydantic_model_creator, pydantic_queryset_creator
from tortoise.expressions import F
from tortoise.transactions import in_transaction

from ayy import tools as tools_module
from ayy.db_models import DEFAULT_APP_NAME
from ayy.db_models import Dialog as DBDialog
from ayy.db_models import Task as DBTask
from ayy.db_models import TaskTool as DBTaskTool
from ayy.dialog import Dialog, DialogToolSignature, Task, TaskTool, Tool
from ayy.func_utils import get_functions_from_module

DEFAULT_DB_NAME = "tasks_db"
TOOL_FIELDS = ["reasoning", "name", "prompt"]


# def db_dialog_tool_to_dialog_tool(
#     db_dialog_tool: BaseModel | dict, tool_fields: list[str] = TOOL_FIELDS
# ) -> DialogTool:
#     if isinstance(db_dialog_tool, BaseModel):
#         db_dialog_tool = db_dialog_tool.model_dump()
#     return DialogTool(
#         id=db_dialog_tool["id"],
#         position=db_dialog_tool["position"],
#         dialog_id=db_dialog_tool["dialog"]["id"],
#         tool=Tool(**{k: v for k, v in db_dialog_tool.items() if k in tool_fields}),
#         used=db_dialog_tool["used"],
#     )


def db_task_tool_to_task_tool(db_task_tool: BaseModel | dict, tool_fields: list[str] = TOOL_FIELDS) -> TaskTool:
    if isinstance(db_task_tool, BaseModel):
        db_task_tool = db_task_tool.model_dump()
    return TaskTool(
        id=db_task_tool["id"],
        position=db_task_tool["position"],
        task_id=db_task_tool["task"]["id"],
        tool=Tool(**{k: v for k, v in db_task_tool.items() if k in tool_fields}),
        used=db_task_tool["used"],
    )


async def init_db(db_names: list[str] | str = DEFAULT_DB_NAME, app_names: list[str] | str = DEFAULT_APP_NAME):
    db_names = [db_names] if isinstance(db_names, str) else db_names
    app_names = [app_names] if isinstance(app_names, str) else app_names
    assert len(db_names) == len(app_names), "Number of db_names and app_names must be the same"
    await Tortoise.init(
        config={
            "connections": {db_name: f"sqlite://{db_name}.sqlite3" for db_name in db_names},
            "apps": {
                app_name: {"models": ["ayy.db_models"], "default_connection": db_name}
                for app_name, db_name in zip(app_names, db_names)
            },
        }
    )
    await Tortoise.generate_schemas()


# async def get_next_dialog_tool(
#     dialog: UUID4 | Dialog,
#     db_name: str = DEFAULT_DB_NAME,
#     used: bool = False,
#     reverse: bool = False,
#     position: int | None = None,
# ) -> DialogTool | None:
#     dialog_id = dialog.id if isinstance(dialog, Dialog) else dialog
#     dialog_tools = DBDialogTool.filter(dialog_id=dialog_id).using_db(connections.get(db_name))
#     if position is not None:
#         dialog_tools = await dialog_tools.filter(position=position).first()
#     else:
#         dialog_tools = (
#             await dialog_tools.filter(used=used).order_by("position" if not reverse else "-position").first()
#         )
#     if dialog_tools is None:
#         return None
#     tool_model = pydantic_model_creator(DBDialogTool)
#     tool_model = await tool_model.from_tortoise_orm(await dialog_tools)
#     return db_dialog_tool_to_dialog_tool(db_dialog_tool=tool_model)


async def get_next_task_tool(
    task: UUID4 | Task,
    db_name: str = DEFAULT_DB_NAME,
    used: bool = False,
    reverse: bool = False,
    position: int | None = None,
) -> TaskTool | None:
    task_id = task.id if isinstance(task, Task) else task
    task_tools = DBTaskTool.filter(task_id=task_id).using_db(connections.get(db_name))
    if position is not None:
        task_tools = await task_tools.filter(position=position).first()
    else:
        task_tools = (
            await task_tools.filter(used=used).order_by("position" if not reverse else "-position").first()
        )
    if task_tools is None:
        return None
    tool_model = pydantic_model_creator(DBTaskTool)
    tool_model = await tool_model.from_tortoise_orm(await task_tools)
    return db_task_tool_to_task_tool(db_task_tool=tool_model)


async def get_dialogs_with_signatures(db_name: str = DEFAULT_DB_NAME) -> list[Dialog]:
    dialogs = await pydantic_queryset_creator(DBDialog).from_queryset(
        DBDialog.filter(dialog_tool_signature__not={}).using_db(connections.get(db_name))
    )
    return [Dialog(**dialog) for dialog in dialogs.model_dump()]


# async def get_dialog_tools(
#     dialog: UUID4 | Dialog, db_name: str = DEFAULT_DB_NAME, used: bool = False
# ) -> list[DialogTool]:
#     dialog_id = dialog.id if isinstance(dialog, Dialog) else dialog
#     dialog_tools = await pydantic_queryset_creator(DBDialogTool).from_queryset(
#         DBDialogTool.filter(dialog_id=dialog_id, used=used).order_by("position").using_db(connections.get(db_name))
#     )
#     return [db_dialog_tool_to_dialog_tool(db_dialog_tool=tool) for tool in dialog_tools.model_dump()]


async def get_task_tools(task: UUID4 | Task, db_name: str = DEFAULT_DB_NAME, used: bool = False) -> list[TaskTool]:
    task_id = task.id if isinstance(task, Task) else task
    task_tools = await pydantic_queryset_creator(DBTaskTool).from_queryset(
        DBTaskTool.filter(task_id=task_id, used=used).order_by("position").using_db(connections.get(db_name))
    )
    return [db_task_tool_to_task_tool(db_task_tool=tool) for tool in task_tools.model_dump()]


# async def add_dialog_tools(
#     dialog: UUID4 | Dialog,
#     tools: list[Tool] | Tool,
#     db_name: str = DEFAULT_DB_NAME,
#     position: int | None = None,
#     used: bool = False,
#     run_next: bool = False,
#     replace_all: bool = False,
#     tool_fields: list[str] = TOOL_FIELDS,
# ) -> None:
#     tools = [tools] if isinstance(tools, Tool) else tools
#     conn = connections.get(db_name)
#     dialog_id = dialog.id if isinstance(dialog, Dialog) else dialog
#     async with in_transaction():
#         conn = connections.get(db_name)
#         dialog_tools = DBDialogTool.filter(dialog_id=dialog_id).using_db(conn)
#         if replace_all:
#             await dialog_tools.delete()
#             start_position = 1
#         else:
#             if run_next:
#                 first_unused = await dialog_tools.filter(used=False).order_by("position").first()
#                 start_position = 1 if first_unused is None else first_unused.position
#                 await dialog_tools.filter(position__gte=start_position).update(position=F("position") + len(tools))
#             else:
#                 query = dialog_tools.order_by("-position")
#                 latest_tool = await query.first()
#                 if latest_tool is None:
#                     latest_position = 0
#                 else:
#                     latest_position = latest_tool.position
#                 logger.info(f"latest_position: {latest_position}")
#                 if position is not None:
#                     await dialog_tools.filter(position__gte=position).update(position=F("position") + len(tools))
#                     start_position = position
#                 else:
#                     start_position = latest_position + 1
#         for i, tool in enumerate(tools, start=start_position):
#             logger.info(f"tool: {tool}")
#             await DBDialogTool.create(
#                 using_db=conn,
#                 dialog_id=dialog_id,
#                 position=i,
#                 used=used,
#                 **tool.model_dump(include=set(tool_fields)),
#             )


async def add_task_tools(
    task: UUID4 | Task,
    tools: list[Tool] | Tool,
    db_name: str = DEFAULT_DB_NAME,
    position: int | None = None,
    used: bool = False,
    run_next: bool = False,
    replace_all: bool = False,
    tool_fields: list[str] = TOOL_FIELDS,
) -> None:
    tools = [tools] if isinstance(tools, Tool) else tools
    conn = connections.get(db_name)
    task_id = task.id if isinstance(task, Task) else task
    async with in_transaction():
        conn = connections.get(db_name)
        task_tools = DBTaskTool.filter(task_id=task_id).using_db(conn)
        if replace_all:
            await task_tools.delete()
            start_position = 1
        else:
            if run_next:
                first_unused = await task_tools.filter(used=False).order_by("position").first()
                start_position = 1 if first_unused is None else first_unused.position
                await task_tools.filter(position__gte=start_position).update(position=F("position") + len(tools))
            else:
                query = task_tools.order_by("-position")
                latest_tool = await query.first()
                if latest_tool is None:
                    latest_position = 0
                else:
                    latest_position = latest_tool.position
                logger.info(f"latest_position: {latest_position}")
                if position is not None:
                    await task_tools.filter(position__gte=position).update(position=F("position") + len(tools))
                    start_position = position
                else:
                    start_position = latest_position + 1
        for i, tool in enumerate(tools, start=start_position):
            logger.info(f"tool: {tool}")
            await DBTaskTool.create(
                using_db=conn, task_id=task_id, position=i, used=used, **tool.model_dump(include=set(tool_fields))
            )


# async def toggle_dialog_tool_usage(dialog_tool_id: int, db_name: str = DEFAULT_DB_NAME) -> None:
#     tool = await DBDialogTool.get(id=dialog_tool_id, using_db=connections.get(db_name))
#     tool.used = not tool.used
#     await tool.save()


async def toggle_task_tool_usage(task_tool_id: int, db_name: str = DEFAULT_DB_NAME) -> None:
    tool = await DBTaskTool.get(id=task_tool_id, using_db=connections.get(db_name))
    tool.used = not tool.used
    await tool.save()


async def save_task(task: Task, db_name: str = DEFAULT_DB_NAME, overwrite: bool = True) -> None:
    conn = connections.get(db_name)
    existing_task = await DBTask.filter(id=task.id).using_db(conn).first()
    if existing_task is None and task.name != "":
        existing_task = await DBTask.filter(name=task.name).using_db(conn).first()
    if existing_task is None:
        await DBTask.create(using_db=conn, **task.model_dump())
    elif overwrite:
        existing_task = await existing_task.update_from_dict(
            {k: v for k, v in task.model_dump().items() if k not in ["id"]}
        )
        await existing_task.save()


async def load_task(task: UUID4 | str | Task, db_name: str = DEFAULT_DB_NAME) -> Task:
    if isinstance(task, Task):
        return task
    conn = connections.get(db_name)
    kwargs = {"name": task} if isinstance(task, str) else {"id": task}
    task_obj, _ = await DBTask.get_or_create(defaults=None, using_db=conn, **kwargs)
    task_model = pydantic_model_creator(DBTask)
    task_model = await task_model.from_tortoise_orm(task_obj)
    return Task(**task_model.model_dump())


async def save_dialog(
    dialog: Dialog,
    # task: Task | None = None,
    db_name: str = DEFAULT_DB_NAME,
    dialog_tool_signature: DialogToolSignature | None = None,
    overwrite: bool = True,
) -> None:
    conn = connections.get(db_name)
    dialog_dict = dialog.model_dump()
    if dialog_tool_signature is not None and dialog_tool_signature.name not in [
        f[0] for f in get_functions_from_module(tools_module)
    ]:
        dialog_dict["system"] = dialog_tool_signature.system
        dialog_dict["name"] = dialog_tool_signature.name
        dialog_dict["dialog_tool_signature"] = dialog_tool_signature.model_dump()
    # if task is not None:
    #     dialog_dict["messages"] += task.messages
    #     await save_task(task=task, db_name=db_name)
    existing_dialog = await DBDialog.filter(name=dialog.name).using_db(conn).first() if dialog.name != "" else None
    if existing_dialog is None:
        existing_dialog = await DBDialog.filter(id=dialog.id).using_db(conn).first()
    if existing_dialog is None:
        await DBDialog.create(using_db=conn, **dialog_dict)
    elif overwrite:
        existing_dialog = await existing_dialog.update_from_dict(
            {k: v for k, v in dialog_dict.items() if k not in ["id", "dialog_id"]}
        )
        await existing_dialog.save()


async def load_dialog(dialog: UUID4 | str | Dialog, db_name: str = DEFAULT_DB_NAME) -> Dialog:
    if isinstance(dialog, Dialog):
        return dialog
    conn = connections.get(db_name)
    kwargs = {"name": dialog} if isinstance(dialog, str) else {"id": dialog}
    dialog_obj, _ = await DBDialog.get_or_create(defaults=None, using_db=conn, **kwargs)
    dialog_model = pydantic_model_creator(DBDialog)
    dialog_model = await dialog_model.from_tortoise_orm(dialog_obj)
    return Dialog(**dialog_model.model_dump())



from enum import StrEnum
from typing import Literal

from pydantic import BaseModel, Field


class MemoryType(StrEnum):
    SEMANTIC = "semantic"
    EPISODIC = "episodic"


class Memory(BaseModel):
    name: str
    content: str
    confidence: float = Field(default=1.0, ge=0.0, le=1.0)


class SemanticCategory(StrEnum):
    IDENTITY = "identity"
    PREFERENCES = "preferences"
    RELATIONSHIPS = "relationships"
    SKILLS = "skills"
    BELIEFS = "beliefs"
    BACKGROUND = "background"
    HEALTH = "health"
    LOCATION = "location"
    SCHEDULE = "schedule"
    GOALS = "goals"
    OTHER = "other"


class SemanticMemory(Memory):
    """Facts and knowledge that persist across conversations"""

    type: Literal[MemoryType.SEMANTIC] = MemoryType.SEMANTIC
    category: SemanticCategory


class EpisodicMemory(Memory):
    """Experiences and events tied to specific contexts"""

    type: Literal[MemoryType.EPISODIC] = MemoryType.EPISODIC
    context: str = Field(description="Context this memory is relevant to")


class Summary(BaseModel):
    bullet_points: list[str]
    semantic_memories: list[SemanticMemory] = Field(default_factory=list)
    episodic_memories: list[EpisodicMemory] = Field(default_factory=list)



import json
from copy import deepcopy
from enum import StrEnum
from functools import partial
from itertools import groupby
from operator import itemgetter
from pathlib import Path
from typing import Annotated, Any, Self, cast
from uuid import uuid4

import instructor
import numpy as np
from anthropic import Anthropic, AsyncAnthropic
from google.generativeai import GenerativeModel
from instructor import AsyncInstructor, Instructor
from loguru import logger
from openai import AsyncOpenAI, OpenAI
from pydantic import UUID4, AfterValidator, BaseModel, Field, field_validator, model_validator

from ayy.memory import Summary
from ayy.prompts import SELECT_TOOLS
from ayy.utils import flatten_list

MAX_MESSAGE_TOKENS = 100_000
MERGE_JOINER = "\n\n--- Next Message ---\n\n"
TEMPERATURE = 0.1
MAX_TOKENS = 3000
DEFAULT_TAG = "RECALL"


class ModelName(StrEnum):
    GPT = "gpt-4o-2024-11-20"
    GPT_MINI = "gpt-4o-mini"
    HAIKU = "claude-3-haiku-latest"
    SONNET = "claude-3-5-sonnet-latest"
    OPUS = "claude-3-opus-latest"
    GEMINI_PRO = "gemini-1.5-pro-001"
    GEMINI_FLASH = "gemini-1.5-flash-002"
    GEMINI_FLASH_EXP = "gemini-1.5-flash-exp-0827"


class Tool(BaseModel):
    id: UUID4 | str = Field(default_factory=lambda: str(uuid4()))
    reasoning: str
    name: str
    prompt: str = Field(
        ...,
        description="An LLM will receive the messages so far and the tools calls and results up until now. This prompt will then be used to ask the LLM to generate arguments for the selected tool based on the tool's signature. If the tool doesn't have any parameters, then it doesn't need a prompt.",
    )

    @field_validator("id")
    @classmethod
    def validate_id(cls, v: UUID4 | str) -> str:
        if isinstance(v, str):
            v = uuid4()
        return str(v)

    def __str__(self) -> str:
        return f"Reasoning: {self.reasoning}\nName: {self.name}\nPrompt: {self.prompt}"


def load_content(content: Any, echo: bool = False) -> Any:
    if not isinstance(content, (str, Path)):
        return content
    else:
        try:
            return Path(content).read_text()
        except Exception as e:
            if echo:
                logger.warning(f"Could not load content as a file: {str(e)[:100]}")
            return str(content)


MODEL_NAME = ModelName.GEMINI_FLASH
Content = Annotated[Any, AfterValidator(load_content)]
MessageType = dict[str, Content]


class MessagePurpose(StrEnum):
    SUMMARY = "summary"
    CONVO = "convo"
    TOOL = "tool"
    ERROR = "error"


DEFAULT_MESSAGE_PURPOSE = MessagePurpose.CONVO


def create_creator(model_name: ModelName = MODEL_NAME, use_async: bool = False) -> Instructor | AsyncInstructor:
    if "gpt" in model_name.lower():
        if use_async:
            client = instructor.from_openai(AsyncOpenAI())
        else:
            client = instructor.from_openai(OpenAI())
    elif "claude" in model_name.lower():
        if use_async:
            client = instructor.from_anthropic(AsyncAnthropic())
        else:
            client = instructor.from_anthropic(Anthropic())
    elif "gemini" in model_name.lower():
        client = instructor.from_gemini(
            client=GenerativeModel(model_name=model_name),
            mode=instructor.Mode.GEMINI_JSON,
            use_async=use_async,  # type: ignore
        )
    else:
        raise ValueError(f"Model {model_name} not supported")
    return client.chat.completions


def chat_message(
    role: str, content: Content, template: Content = "", purpose: MessagePurpose = DEFAULT_MESSAGE_PURPOSE
) -> MessageType:
    if template:
        if not isinstance(content, dict):
            raise TypeError("When using template, content must be a dict.")
        try:
            message_content = template.format(**content)
        except KeyError as e:
            raise KeyError(f"Template {template} requires key {e} which was not found in content.")
    else:
        message_content = content
    return {"role": role, "content": message_content, "purpose": purpose}


def system_message(
    content: Content, template: Content = "", purpose: MessagePurpose = DEFAULT_MESSAGE_PURPOSE
) -> MessageType:
    return chat_message(role="system", content=content, template=template, purpose=purpose)


def user_message(
    content: Content, template: Content = "", purpose: MessagePurpose = DEFAULT_MESSAGE_PURPOSE
) -> MessageType:
    return chat_message(role="user", content=content, template=template, purpose=purpose)


def assistant_message(
    content: Content, template: Content = "", purpose: MessagePurpose = DEFAULT_MESSAGE_PURPOSE
) -> MessageType:
    return chat_message(role="assistant", content=content, template=template, purpose=purpose)


def load_messages(messages: list[MessageType] | str | Path) -> list[MessageType]:
    if isinstance(messages, list):
        return messages
    else:
        try:
            return json.loads(Path(messages).read_text())
        except Exception as e:
            logger.warning(f"Could not load messages as a file: {str(e)[:100]}")
            return [user_message(content=str(messages))]


Messages = Annotated[list[MessageType], AfterValidator(load_messages)]


class DialogToolSignature(BaseModel):
    name: str = ""
    signature: str = ""
    docstring: str = ""
    system: str = ""


class Dialog(BaseModel):
    id: UUID4 = Field(default_factory=lambda: uuid4())
    name: str = ""
    system: Content = ""
    messages: Messages = Field(default_factory=list)
    model_name: ModelName = MODEL_NAME
    creation_config: dict = dict(temperature=TEMPERATURE, max_tokens=MAX_TOKENS)
    max_message_tokens: int = MAX_MESSAGE_TOKENS
    dialog_tool_signature: dict = Field(default_factory=dict)
    available_tools: list[str] = Field(default_factory=list)
    include_tool_guidelines: bool = True

    @model_validator(mode="after")
    def validate_system_and_signature(self) -> Self:
        self.system = self.dialog_tool_signature.get("system", self.system).strip()
        if self.include_tool_guidelines:
            self.system += f"\n\n<tool_selection_guidelines>\n{SELECT_TOOLS}\n</tool_selection_guidelines>"
        self.system = self.system.strip()
        if "system" in self.dialog_tool_signature:
            del self.dialog_tool_signature["system"]
        return self

    @field_validator("max_message_tokens")
    @classmethod
    def validate_max_message_tokens(cls, v: int) -> int:
        if "gpt" in cls.model_name.lower():
            return 100_000
        elif "claude" in cls.model_name.lower():
            return 175_000
        elif "gemini" in cls.model_name.lower():
            return 800_000
        return v


class Task(BaseModel):
    id: UUID4 = Field(default_factory=lambda: uuid4())
    name: str = ""
    dialog_id: UUID4
    messages: Messages = Field(default_factory=list)


class TaskTool(BaseModel):
    id: int
    position: int
    task_id: UUID4
    tool: Tool
    used: bool = False


def get_last_message(messages: Messages, role: str = "assistant") -> MessageType | None:
    return next((msg for msg in reversed(messages) if msg["role"] == role), None)


def exchange(
    user: Content,
    assistant: Content,
    feedback: Content = "",
    correction: Content = "",
    user_template: Content = "",
    assistant_template: Content = "",
) -> list[MessageType]:
    user_maker = partial(user_message, template=user_template)
    assistant_maker = partial(assistant_message, template=assistant_template)
    return (
        [user_maker(content=user), assistant_maker(content=assistant)]
        + ([user_maker(content=feedback)] if feedback else [])
        + ([assistant_maker(content=correction)] if correction else [])
    )


def merge_same_role_messages(messages: Messages, joiner: Content = MERGE_JOINER) -> list[MessageType]:
    return (
        [
            {"role": role, "content": joiner.join(msg["content"] for msg in group)}
            for role, group in groupby(messages, key=itemgetter("role"))
        ]
        if messages
        else []
    )


def count_tokens(messages: Messages) -> int:
    return int(np.ceil(len(flatten_list([message["content"].split() for message in messages])) * 0.75))


def trim_messages(
    messages: Messages, max_message_tokens: int | None = MAX_MESSAGE_TOKENS
) -> tuple[list[MessageType], list[MessageType]]:
    if not messages or max_message_tokens is None:
        return messages, []
    system_msg = [messages[0]] if messages and messages[0]["role"] == "system" else []
    working_messages = messages[1:] if system_msg else messages[:]
    if working_messages and working_messages[0]["role"] == "assistant":
        working_messages[0] = {
            **working_messages[0],
            "role": "user",
            "content": f"<previous_assistant_message>\n{working_messages[0]['content']}\n</previous_assistant_message>",
        }
    total_tokens = 0
    trim_idx = len(working_messages) - 1
    while trim_idx >= 0:
        msg_tokens = count_tokens([working_messages[trim_idx]])
        if total_tokens + msg_tokens > max_message_tokens:
            break
        total_tokens += msg_tokens
        trim_idx -= 1
    return system_msg + working_messages[trim_idx + 1 :], working_messages[: trim_idx + 1]


def messages_to_kwargs(
    messages: Messages,
    system: str = "",
    model_name: str = MODEL_NAME,
    joiner: Content = MERGE_JOINER,
    max_message_tokens: int = MAX_MESSAGE_TOKENS,
) -> dict:
    messages = deepcopy(messages)
    messages, _ = trim_messages(messages=messages, max_message_tokens=max_message_tokens)
    kwargs = {"messages": [chat_message(role=message["role"], content=message["content"]) for message in messages]}
    first_message = messages[0]
    if first_message["role"] == "system":
        system = system or first_message["content"]
        kwargs["messages"][0]["content"] = system
    else:
        kwargs["messages"].insert(0, system_message(content=system))
    if any(name in model_name.lower() for name in ("gemini", "claude")):
        kwargs["messages"] = merge_same_role_messages(messages=kwargs["messages"], joiner=joiner)
    if "claude" in model_name.lower():
        return {"system": system, "messages": kwargs["messages"][1:]}
    return kwargs


def dialog_to_kwargs(dialog: Dialog, messages: Messages | None = None) -> dict:
    messages = messages or []
    kwargs = messages_to_kwargs(
        messages=dialog.messages + [message for message in messages if message["role"] != "system"],
        system=dialog.system,
        model_name=dialog.model_name,
        max_message_tokens=dialog.max_message_tokens,
    )
    if "gemini" in dialog.model_name.lower():
        kwargs["generation_config"] = dialog.creation_config
    else:
        kwargs.update(dialog.creation_config)
    return kwargs


def summarize_messages(
    messages: Messages, summarizer_dialog: Dialog | None = None, max_message_tokens: int = MAX_MESSAGE_TOKENS
) -> tuple[Messages, Summary | None]:
    if summarizer_dialog is None:
        return messages, None
    messages = deepcopy(messages)
    messages, trimmed_messages = trim_messages(messages=messages, max_message_tokens=max_message_tokens)
    if not trimmed_messages:
        return messages, None
    try:
        creator = create_creator(model_name=summarizer_dialog.model_name)
        summary: Summary = creator.create(
            **dialog_to_kwargs(
                dialog=summarizer_dialog,
                messages=trimmed_messages + [user_message(content="Summarize the conversation so far.")],
            ),
            response_model=Summary,  # type: ignore
        )
        summary_message = user_message(
            content=f"<summary_of_our_previous_conversation(s)>\n{json.dumps(summary.model_dump(), indent=2)}\n</summary_of_our_previous_conversation(s)>"
        )
        if messages[0]["role"] == "system":
            messages = [messages[0], summary_message] + messages[1:]
        else:
            messages = [summary_message] + messages
        return messages, summary
    except Exception:
        logger.exception(f"Could not summarize messages: {trimmed_messages[-2:]}")
        return messages, None


def add_message(
    task_or_dialog: Task | Dialog,
    role: str = "user",
    content: Content = "",
    template: Content = "",
    purpose: MessagePurpose = DEFAULT_MESSAGE_PURPOSE,
    message: MessageType | None = None,
    summarizer_dialog: Dialog | None = None,
) -> Task | Dialog:
    if content is None:
        return task_or_dialog
    task_or_dialog.messages = summarize_messages(
        messages=task_or_dialog.messages
        + [message or chat_message(role=role, content=content, template=template, purpose=purpose)],
        summarizer_dialog=summarizer_dialog,
    )[0]
    return task_or_dialog


def add_dialog_message(
    dialog: Dialog,
    role: str = "user",
    content: Content = "",
    template: Content = "",
    purpose: MessagePurpose = DEFAULT_MESSAGE_PURPOSE,
    message: MessageType | None = None,
    summarizer_dialog: Dialog | None = None,
) -> Dialog:
    return cast(
        Dialog,
        add_message(
            task_or_dialog=dialog,
            role=role,
            content=content,
            template=template,
            purpose=purpose,
            message=message,
            summarizer_dialog=summarizer_dialog,
        ),
    )


def add_task_message(
    task: Task,
    role: str = "user",
    content: Content = "",
    template: Content = "",
    purpose: MessagePurpose = DEFAULT_MESSAGE_PURPOSE,
    message: MessageType | None = None,
) -> Task:
    return cast(
        Task,
        add_message(
            task_or_dialog=task, role=role, content=content, template=template, purpose=purpose, message=message
        ),
    )



import inspect
import typing
from functools import partial
from inspect import Parameter, Signature
from typing import Any, Callable

from pydantic import create_model

from ayy.utils import deindent


def get_functions_from_module(
    module, exclude: list[str] | set[str] | None = None, include: list[str] | set[str] | None = None
):
    exclude = set(exclude) if exclude else set()
    include = set(include) if include else set()
    return [
        x
        for x in inspect.getmembers(
            module, lambda member: inspect.isfunction(member) and member.__module__ == module.__name__
        )
        if not x[0].startswith("_") and x[0] not in exclude and (x[0] in include if include else True)
    ]


def get_param_names(func: Callable):
    func = func.func if isinstance(func, partial) else func
    return inspect.signature(func).parameters.keys()


def get_required_param_names(func: Callable) -> list[str]:
    if isinstance(func, partial):
        params = inspect.signature(func.func).parameters
        return [
            name
            for name, param in params.items()
            if param.default == Parameter.empty and name not in func.keywords.keys()
        ]
    params = inspect.signature(func).parameters
    return [name for name, param in params.items() if param.default == Parameter.empty]


def function_schema(func: Callable) -> dict:
    kw = {
        n: (o.annotation, ... if o.default == Parameter.empty else o.default)
        for n, o in inspect.signature(func).parameters.items()
    }
    s = create_model(f"Input for `{func.__name__}`", **kw).schema()  # type: ignore
    return dict(name=func.__name__, description=func.__doc__, parameters=s)


def skip_params(
    func: Callable, skip_default_params: bool = False, params_to_skip: list[str] | None = None
) -> Signature:
    params_to_skip = params_to_skip or []
    if isinstance(func, partial):
        signature = inspect.signature(func.func)
        print(f"Keywords: {func.keywords}")
        return signature.replace(
            parameters=[
                p
                for p in signature.parameters.values()
                if (p.default == Parameter.empty and p.name not in func.keywords and skip_default_params)
                or (p.name in params_to_skip)
            ]
        )
    else:
        signature = inspect.signature(func)
        return signature.replace(
            parameters=[
                p
                for p in signature.parameters.values()
                if (p.default == Parameter.empty and skip_default_params) or (p.name in params_to_skip)
            ]
        )


def get_function_signature(
    func: Callable,
    ignore_default_values: bool = False,
    skip_default_params: bool = False,
    params_to_skip: list[str] | None = None,
) -> Signature:
    if skip_default_params or params_to_skip:
        return skip_params(func, skip_default_params=skip_default_params, params_to_skip=params_to_skip)
    else:
        if isinstance(func, partial):
            signature = inspect.signature(func.func)
        else:
            signature = inspect.signature(func)
        if ignore_default_values:
            return signature.replace(
                parameters=[
                    Parameter(name=p.name, kind=p.kind, default=Parameter.empty, annotation=p.annotation)
                    for p in signature.parameters.values()
                ]
            )
        else:
            return signature


def function_to_type(
    func: Callable | type,
    ignore_default_values: bool = False,
    skip_default_params: bool = False,
    params_to_skip: list[str] | None = None,
) -> type:
    if isinstance(func, type):
        return func
    kw = {
        n: (
            str if o.annotation == Parameter.empty else o.annotation,
            ... if (o.default == Parameter.empty or ignore_default_values) else o.default,
        )
        for n, o in (
            get_function_signature(
                func,
                ignore_default_values=ignore_default_values,
                skip_default_params=skip_default_params,
                params_to_skip=params_to_skip,
            )
        ).parameters.items()
    }
    if isinstance(func, partial):
        func = func.func
    return create_model(func.__name__, __doc__=inspect.getdoc(func), **kw)  # type:ignore


def get_function_return_type(func: Callable) -> type:
    func = func.func if isinstance(func, partial) else func
    sig = typing.get_type_hints(func)
    return sig.get("return", None)


def get_function_name(func: Callable) -> str:
    func = func.func if isinstance(func, partial) else func
    return func.__name__


def get_function_source(func: Callable) -> str:
    func = func.func if isinstance(func, partial) else func
    return inspect.getsource(func)


def get_function_info(
    func: Callable, ignore_default_values: bool = False, skip_default_params: bool = False
) -> dict[str, str]:
    signature = get_function_signature(
        func=func, ignore_default_values=ignore_default_values, skip_default_params=skip_default_params
    )
    func = func.func if isinstance(func, partial) else func
    name = func.__name__
    docstring = inspect.getdoc(func)
    info = {"name": name}
    if signature:
        info["signature"] = f"{name}{signature}"
    if docstring:
        info["docstring"] = deindent(docstring)
    return info


def create_function_info(
    name: str, parameters: dict[str, tuple[type, Any]] | None = None, docstring: str = ""
) -> dict[str, str]:
    parameters = parameters or {}
    sig_params = []
    for param_name, (param_type, default_value) in parameters.items():
        sig_params.append(
            Parameter(
                name=param_name,
                annotation=param_type,
                default=Parameter.empty if default_value is ... else default_value,
                kind=Parameter.POSITIONAL_OR_KEYWORD,
            )
        )
    signature = Signature(parameters=sig_params)
    info = {"name": name}
    info["signature"] = f"{name}{signature}".replace("__main__.", "")
    if docstring:
        info["docstring"] = docstring
    return info


def type_to_function_info(typ: type, func_name: str, param_name: str, docstring: str) -> dict[str, str]:
    info = {"name": func_name}
    info["signature"] = f"{func_name}({param_name}: {typ.__name__})"
    if docstring:
        info["docstring"] = docstring
    return info



from loguru import logger
from tortoise import run_async

from ayy.dialog import Dialog, ModelName
from ayy.leggo import new_task
from ayy.torm import init_db, save_dialog

MODEL_NAME = ModelName.GEMINI_FLASH


DB_NAME = "tasks_db"
APP_NAME = "tasks"


async def setup():
    await init_db(db_names=DB_NAME, app_names=APP_NAME)
    await save_dialog(
        dialog=Dialog(model_name=MODEL_NAME, name="default_dialog"), db_name=DB_NAME, overwrite=False
    )


if __name__ == "__main__":
    logger.info("Setting up")
    run_async(setup())
    logger.success("Setup done")
    logger.info("Running task")
    run_async(
        new_task(
            db_name=DB_NAME,
            dialog="list_grounds",
            # task_query="list the grounds in london",
            # task_query="list the grounds in manchester and the weather there on tue",
            task_query="weather on tuesday?",
        )
    )
    logger.success("Task done")



import inspect
import re
import textwrap
from itertools import chain
from pathlib import Path
from typing import Iterable


def flatten(o: Iterable):
    for item in o:
        if isinstance(item, str):
            yield item
            continue
        try:
            yield from flatten(item)
        except TypeError:
            yield item


def to_camel(s: str, sep: str = "_") -> str:
    if sep not in s:
        return s
    return "".join(s.title().split(sep))


def to_snake(s: str) -> str:
    return re.sub(r"(?<!^)(?=[A-Z])", "_", s).lower()


def noop(x=None, *args, **kwargs):  # noqa
    return x


def resolve_data_path(data_path: list[str | Path] | str | Path, file_extension: str | None = None) -> chain:
    if not isinstance(data_path, list):
        data_path = [data_path]
    paths = []
    for dp in flatten(data_path):
        if isinstance(dp, (str, Path)):
            dp = Path(dp)
            if not dp.exists():
                raise Exception(f"Path {dp} does not exist.")
            if dp.is_dir():
                if file_extension:
                    paths.append(dp.glob(f"*.{file_extension}"))
                else:
                    paths.append(dp.iterdir())
            else:
                if file_extension is None or dp.suffix == f".{file_extension}":
                    paths.append([dp])
    return chain(*paths)


def flatten_list(my_list: list) -> list:
    new_list = []
    for x in my_list:
        if isinstance(x, list):
            new_list += flatten_list(x)
        else:
            new_list.append(x)
    return new_list


def deindent(text: str) -> str:
    return textwrap.dedent(inspect.cleandoc(text))


def remove_digits(text: str) -> str:
    return re.sub(r"\d+", "", text)
